// (c) by Stefan Roettger, licensed under LGPL 2.1

#include "minibase.h"

#include "miniARB.h"

int miniARB::INSTANCES=0;

int miniARB::USEGLOBALSHADER=-1;

miniARB::SHADER_TYPE miniARB::SHADER[SHADERMAX];

miniARB::SNIPPET_TYPE miniARB::SNIPPET[SNIPPETMAX];
int miniARB::SNIPPETS=0;

int miniARB::global_shader[256]={-1};

BOOLINT miniARB::global_texgen=FALSE;
BOOLINT miniARB::global_shade=FALSE,miniARB::global_shade_direct=FALSE;
BOOLINT miniARB::global_tex=FALSE,miniARB::global_tex3=FALSE;
BOOLINT miniARB::global_fog=FALSE;
BOOLINT miniARB::global_invariant=FALSE;
BOOLINT miniARB::global_sfx=FALSE;

float miniARB::global_fogstart=0.0f,miniARB::global_fogend=0.0f;
float miniARB::global_fogdensity=0.0f,miniARB::global_fogcolor[3]={0,0,0};

float miniARB::global_lightdir[3]={0,0,0};
float miniARB::global_lightbias=0.5f,miniARB::global_lightoffset=0.5f;
float miniARB::global_transbias=4.0f,miniARB::global_transoffset=0.0f;

int miniARB::global_sfxmode=3;

// initialize shader snippets
void miniARB::initsnippets()
   {
   // vertex program snippets:

   addsnippet(MINI_SNIPPET_VTX_BEGIN,"!!ARBvp1.0\n");

   addsnippet(MINI_SNIPPET_VTX_BEGIN_INVARIANT,"\
      OPTION ARB_position_invariant; \n");

   addsnippet(MINI_SNIPPET_VTX_HEADER,"\
      PARAM c0=program.env[0]; \n\
      PARAM c1=program.env[1]; \n\
      PARAM c2=program.env[2]; \n\
      PARAM c3=program.env[3]; \n\
      PARAM c4=program.env[4]; \n\
      PARAM c5=program.env[5]; \n\
      PARAM c6=program.env[6]; \n\
      PARAM c7=program.env[7]; \n\
      PARAM mat[4]={state.matrix.mvp}; \n\
      PARAM matrix[4]={state.matrix.modelview}; \n\
      PARAM invtra[4]={state.matrix.modelview.invtrans}; \n\
      PARAM texmat[4]={state.matrix.texture[0]}; \n\
      TEMP vtx,col,pos; \n");

   addsnippet(MINI_SNIPPET_VTX_BASIC,"\
      ### fetch actual vertex \n\
      MOV vtx,vertex.position; \n\
      MOV col,vertex.color; \n\
      ### transform vertex with combined modelview \n\
      DP4 pos.x,mat[0],vtx; \n\
      DP4 pos.y,mat[1],vtx; \n\
      DP4 pos.z,mat[2],vtx; \n\
      DP4 pos.w,mat[3],vtx; \n\
      ### write transformed vertex \n\
      MOV result.position,pos; \n");

   addsnippet(MINI_SNIPPET_VTX_BASIC_INVARIANT,"\
      ### fetch actual vertex \n\
      MOV vtx,vertex.position; \n\
      MOV col,vertex.color; \n");

   addsnippet(MINI_SNIPPET_VTX_VIEWPOS,"\
      ### transform vertex with modelview \n\
      DP4 pos.x,matrix[0],vtx; \n\
      DP4 pos.y,matrix[1],vtx; \n\
      DP4 pos.z,matrix[2],vtx; \n\
      DP4 pos.w,matrix[3],vtx; \n\
      ### write view position to tex coords \n\
      MOV result.texcoord[1],pos; \n");

   addsnippet(MINI_SNIPPET_VTX_NORMAL,"\
      ### fetch actual normal \n\
      TEMP nrm; \n\
      MOV nrm,vertex.normal; \n\
      ### transform normal with inverse transpose \n\
      TEMP vec; \n\
      DP4 vec.x,invtra[0],nrm; \n\
      DP4 vec.y,invtra[1],nrm; \n\
      DP4 vec.z,invtra[2],nrm; \n\
      DP4 vec.w,invtra[3],nrm; \n\
      ### write normal to tex coords \n\
      MOV result.texcoord[2],vec; \n");

   addsnippet(MINI_SNIPPET_VTX_NORMAL_DIRECT,"\
      ### directly write normal to tex coords\n\
      MOV result.texcoord[2],vertex.normal; \n");

   addsnippet(MINI_SNIPPET_VTX_TEX,"\
      ### fetch actual tex coords \n\
      TEMP texcrd; \n\
      MOV texcrd,vertex.texcoord[0]; \n\
      ### transform tex coords with texture matrix \n\
      TEMP crd; \n\
      DP4 crd.x,texmat[0],texcrd; \n\
      DP4 crd.y,texmat[1],texcrd; \n\
      DP4 crd.z,texmat[2],texcrd; \n\
      DP4 crd.w,texmat[3],texcrd; \n\
      ### write resulting tex coords \n\
      MOV result.texcoord[0],crd; \n");

   addsnippet(MINI_SNIPPET_VTX_TEXGEN,"\
      ### generate tex coords from vertex position \n\
      TEMP crd; \n\
      DP4 crd.x,texmat[0],vtx; \n\
      DP4 crd.y,texmat[1],vtx; \n\
      DP4 crd.z,texmat[2],vtx; \n\
      DP4 crd.w,texmat[3],vtx; \n\
      ### write resulting tex coords \n\
      MOV result.texcoord[0],crd; \n");

   addsnippet(MINI_SNIPPET_VTX_FOG,"\
      ### calculate spherical fog coord \n\
      DP3 result.fogcoord.x,pos,pos; \n");

   addsnippet(MINI_SNIPPET_VTX_FOOTER,"\
      ### write resulting color \n\
      MOV result.color,col; \n");

   addsnippet(MINI_SNIPPET_VTX_END,"END\n");

   // fragment program snippets:

   addsnippet(MINI_SNIPPET_FRG_BEGIN,"!!ARBfp1.0\n");

   addsnippet(MINI_SNIPPET_FRG_HEADER,"\
      PARAM c0=program.env[0]; \n\
      PARAM c1=program.env[1]; \n\
      PARAM c2=program.env[2]; \n\
      PARAM c3=program.env[3]; \n\
      PARAM c4=program.env[4]; \n\
      PARAM c5=program.env[5]; \n\
      PARAM c6=program.env[6]; \n\
      PARAM c7=program.env[7]; \n\
      TEMP col; \n");

   addsnippet(MINI_SNIPPET_FRG_INTERLACE,"\
      ### stereo interlacing \n\
      TEMP tmp; \n\
      MAD tmp.xy,fragment.position,c2,c2.zwxy; \n\
      FRC tmp.xy,tmp; \n\
      SUB tmp.xy,tmp,0.5; \n\
      KIL tmp.xyxy; \n");

   addsnippet(MINI_SNIPPET_FRG_BASIC,"\
      ### fetch fragment color \n\
      MOV col,fragment.color; \n");

   addsnippet(MINI_SNIPPET_FRG_TEX,"\
      ### fetch actual texel \n\
      TEMP tex; \n\
      TEX tex,fragment.texcoord[0],texture[0],2D; \n\
      ### modulate with texture color \n\
      MUL col,col,tex; \n");

   addsnippet(MINI_SNIPPET_FRG_TEX3,"\
      ### fetch actual voxel \n\
      TEMP tex; \n\
      TEX tex,fragment.texcoord[0],texture[0],3D; \n\
      ### modulate with texture color \n\
      MUL col,col,tex; \n");

   addsnippet(MINI_SNIPPET_FRG_TEX_MASK,"\
      ### fetch actual texel \n\
      TEMP tex,crd; \n\
      TEX tex,fragment.texcoord[0],texture[0],2D; \n\
      ### check coordinate range \n\
      MOV crd,fragment.texcoord[0]; \n\
      CMP tex.a,crd.x,0.0,tex.a; \n\
      CMP tex.a,crd.y,0.0,tex.a; \n\
      SUB crd,1.0,crd; \n\
      CMP tex.a,crd.x,0.0,tex.a; \n\
      CMP tex.a,crd.y,0.0,tex.a; \n\
      ### modulate with texture color \n\
      MUL col,col,tex; \n");

   addsnippet(MINI_SNIPPET_FRG_TEX2_DIRECT,"\
      ### fetch two texels \n\
      TEMP tex1,tex2; \n\
      TEX tex1,fragment.texcoord[0],texture[0],2D; \n\
      TEX tex2,fragment.texcoord[0],texture[1],2D; \n\
      ### fetch fragment normal \n\
      TEMP nrm; \n\
      MOV nrm,fragment.texcoord[2]; \n\
      ### calculate directional light \n\
      TEMP len; \n\
      DP3 len.x,nrm,nrm; \n\
      RSQ len.x,len.x; \n\
      MUL nrm,nrm,len.x; \n\
      DP3 nrm.z,nrm,c5; \n\
      ### blend between texels \n\
      MAD nrm.z,nrm.z,c4.x,c4.y; \n\
      MAD_SAT nrm.z,nrm.z,c4.w,c4.w; \n\
      MUL nrm.z,nrm.z,c5.w; \n\
      SIN nrm.z,nrm.z; \n\
      LRP tex1,nrm.z,tex1,tex2; \n\
      ### modulate with texture color \n\
      MUL col,col,tex1; \n");

   addsnippet(MINI_SNIPPET_FRG_SHADE,"\
      ### fetch view position \n\
      TEMP pos; \n\
      MOV pos,fragment.texcoord[1]; \n\
      ### fetch fragment normal \n\
      TEMP nrm; \n\
      MOV nrm,fragment.texcoord[2]; \n\
      ### calculate head light \n\
      TEMP len; \n\
      DP3 len.x,nrm,nrm; \n\
      RSQ len.x,len.x; \n\
      MUL nrm,nrm,len.x; \n\
      DP3 len.x,pos,pos; \n\
      RSQ len.x,len.x; \n\
      MUL pos,pos,len.x; \n\
      DP3 nrm.z,nrm,pos; \n\
      ### modulate fragment color \n\
      ABS nrm.z,nrm.z; \n\
      MUL col.xyz,col,nrm.z; \n");

   addsnippet(MINI_SNIPPET_FRG_SHADE_ALPHA,"\
      ### fetch view position \n\
      TEMP pos; \n\
      MOV pos,fragment.texcoord[1]; \n\
      ### fetch fragment normal \n\
      TEMP nrm; \n\
      MOV nrm,fragment.texcoord[2]; \n\
      ### calculate head light \n\
      TEMP len; \n\
      DP3 len.x,nrm,nrm; \n\
      RSQ len.x,len.x; \n\
      MUL nrm,nrm,len.x; \n\
      DP3 len.x,pos,pos; \n\
      RSQ len.x,len.x; \n\
      MUL pos,pos,len.x; \n\
      DP3 nrm.z,nrm,pos; \n\
      ### modulate fragment alpha \n\
      ABS nrm.z,nrm.z; \n\
      MUL col.a,col.a,nrm.z; \n");

   addsnippet(MINI_SNIPPET_FRG_SHADE_ALPHA2,"\
      ### fetch view position \n\
      TEMP pos; \n\
      MOV pos,fragment.texcoord[1]; \n\
      ### fetch fragment normal \n\
      TEMP nrm; \n\
      MOV nrm,fragment.texcoord[2]; \n\
      ### calculate head light \n\
      TEMP len; \n\
      DP3 len.x,nrm,nrm; \n\
      RSQ len.x,len.x; \n\
      MUL nrm,nrm,len.x; \n\
      DP3 len.x,pos,pos; \n\
      RSQ len.x,len.x; \n\
      MUL pos,pos,len.x; \n\
      DP3 nrm.z,nrm,pos; \n\
      ### modulate fragment alpha \n\
      MUL nrm.z,nrm.z,1.5; \n\
      MUL nrm.z,nrm.z,nrm.z; \n\
      MUL col.a,col.a,nrm.z; \n");

   addsnippet(MINI_SNIPPET_FRG_SHADE_DIRECT,"\
      ### fetch fragment normal \n\
      TEMP nrm; \n\
      MOV nrm,fragment.texcoord[2]; \n\
      ### calculate directional light \n\
      TEMP len; \n\
      DP3 len.x,nrm,nrm; \n\
      RSQ len.x,len.x; \n\
      MUL nrm,nrm,len.x; \n\
      DP3_SAT nrm.z,nrm,c5; \n\
      ### modulate fragment color \n\
      MAD nrm.z,nrm.z,c4.x,c4.y; \n\
      MUL col.xyz,col,nrm.z; \n");

   addsnippet(MINI_SNIPPET_FRG_FOG,"\
      ### fetch fog coord \n\
      TEMP fog; \n\
      MOV fog.x,fragment.fogcoord.x; \n\
      ### modulate with spherical fog \n\
      MAD_SAT fog.x,fog.x,c6.x,c6.y; \n\
      POW fog.x,fog.x,c6.z; \n\
      LRP col.xyz,fog.x,c7,col; \n");

   addsnippet(MINI_SNIPPET_FRG_FOOTER,"\
      ### write resulting pixel \n\
      MOV result.color,col; \n");

   addsnippet(MINI_SNIPPET_FRG_END,"END\n");
   }

// free shader snippets
void miniARB::freesnippets()
   {
   int n;

   for (n=0; n<SNIPPETS; n++)
      {
      if (SNIPPET[n].snippetname!=NULL) free(SNIPPET[n].snippetname);
      if (SNIPPET[n].snippet!=NULL) free(SNIPPET[n].snippet);
      }

   SNIPPETS=0;
   }

// add shader snippet
void miniARB::addsnippet(const char *snippetname,const char *snippet)
   {
   if (SNIPPETS>=SNIPPETMAX) ERRORMSG();

   SNIPPET[SNIPPETS].snippetname=strdup(snippetname);
   SNIPPET[SNIPPETS].snippet=strdup(snippet);

   SNIPPETS++;
   }

// get free shader slot
int miniARB::getfreeslot()
   {
   int n;

   for (n=0; n<SHADERMAX; n++)
      if (SHADER[n].occupied==0)
         {
         SHADER[n].occupied=1;
         return(n);
         }

   return(-1);
   }

// initialize default shader
void miniARB::initshader()
   {
   int slot;

   for (slot=0; slot<32; slot++)
      global_shader[slot]=createshader(slot&1,
                                       slot&2,slot&4,
                                       slot&8,slot&16);

   for (slot=32; slot<256; slot++) global_shader[slot]=-1;
   }

// create basic shader
int miniARB::createshader(BOOLINT texgen,
                            BOOLINT shade,BOOLINT shade_direct,
                            BOOLINT tex,BOOLINT tex3,
                            BOOLINT fog,
                            BOOLINT invariant,
                            BOOLINT sfx)
   {
   int slot=getfreeslot();

   concatvtxshader(slot,MINI_SNIPPET_VTX_BEGIN);
   if (invariant)
      concatvtxshader(slot,MINI_SNIPPET_VTX_BEGIN_INVARIANT);
   concatvtxshader(slot,MINI_SNIPPET_VTX_HEADER);
   if (!invariant)
      concatvtxshader(slot,MINI_SNIPPET_VTX_BASIC);
   else
      concatvtxshader(slot,MINI_SNIPPET_VTX_BASIC_INVARIANT);
   concatvtxshader(slot,MINI_SNIPPET_VTX_VIEWPOS);
   concatvtxshader(slot,MINI_SNIPPET_VTX_NORMAL);
   if (tex || tex3)
      if (texgen)
         concatvtxshader(slot,MINI_SNIPPET_VTX_TEXGEN);
      else
         concatvtxshader(slot,MINI_SNIPPET_VTX_TEX);
   if (fog)
      concatvtxshader(slot,MINI_SNIPPET_VTX_FOG);
   concatvtxshader(slot,MINI_SNIPPET_VTX_FOOTER);
   concatvtxshader(slot,MINI_SNIPPET_VTX_END);

   concatpixshader(slot,MINI_SNIPPET_FRG_BEGIN);
   concatpixshader(slot,MINI_SNIPPET_FRG_HEADER);
   if (sfx)
      concatpixshader(slot,MINI_SNIPPET_FRG_INTERLACE);
   concatpixshader(slot,MINI_SNIPPET_FRG_BASIC);
   if (tex)
      concatpixshader(slot,MINI_SNIPPET_FRG_TEX);
   else if (tex3)
      concatpixshader(slot,MINI_SNIPPET_FRG_TEX3);
   if (shade)
      concatpixshader(slot,MINI_SNIPPET_FRG_SHADE);
   else if (shade_direct)
      concatpixshader(slot,MINI_SNIPPET_FRG_SHADE_DIRECT);
   if (fog)
      concatpixshader(slot,MINI_SNIPPET_FRG_FOG);
   concatpixshader(slot,MINI_SNIPPET_FRG_FOOTER);
   concatpixshader(slot,MINI_SNIPPET_FRG_END);

   return(slot);
   }

// enable global shader
void miniARB::enableglobalshader(BOOLINT texgen,
                                 BOOLINT shade,BOOLINT shade_direct,
                                 BOOLINT tex,BOOLINT tex3,
                                 BOOLINT fog,
                                 BOOLINT invariant,
                                 BOOLINT sfx)
   {
   int slot;

   slot=texgen+
        2*shade+
        4*shade_direct+
        8*tex+
        16*tex3+
        32*fog+
        64*invariant+
        128*sfx;

   if (global_shader[slot]==-1)
      global_shader[slot]=createshader(texgen,
                                       shade,shade_direct,
                                       tex,tex3,
                                       fog,
                                       invariant,
                                       sfx);

   useglobalshader(global_shader[slot]);
   }

// enable global shader
void miniARB::enableglobalshader()
   {
   enableglobalshader(global_texgen,
                      global_shade,global_shade_direct,
                      global_tex,global_tex3,
                      global_fog,
                      global_invariant,
                      global_sfx);

   settexturedirectparams(getglobalshader(),
                          global_lightdir,
                          global_transbias,global_transoffset);

   setshadedirectparams(getglobalshader(),
                        global_lightdir,
                        global_lightbias,global_lightoffset);

   setfogparams(getglobalshader(),
                global_fogstart,global_fogend,
                global_fogdensity,
                global_fogcolor);

   setsfxparams(getglobalshader(),
                global_sfxmode);
   }

// disable global shader
void miniARB::disableglobalshader()
   {useglobalshader(-1);}

// default constructor
miniARB::miniARB()
   {
   int i,j;

   if (INSTANCES==0)
      {
      for (i=0; i<SHADERMAX; i++)
         {
         SHADER[i].occupied=0;

         SHADER[i].vtxprog=NULL;
         SHADER[i].vtxprogid=0;
         SHADER[i].vtxdirty=0;

         SHADER[i].frgprog=NULL;
         SHADER[i].frgprogid=0;
         SHADER[i].frgdirty=0;

         for (j=0; j<SHADERVTXPRMMAX; j++)
            {
            SHADER[i].vtxshaderpar1[j]=0.0f;
            SHADER[i].vtxshaderpar2[j]=0.0f;
            SHADER[i].vtxshaderpar3[j]=0.0f;
            SHADER[i].vtxshaderpar4[j]=0.0f;
            }

         for (j=0; j<SHADERFRGPRMMAX; j++)
            {
            SHADER[i].pixshaderpar1[j]=0.0f;
            SHADER[i].pixshaderpar2[j]=0.0f;
            SHADER[i].pixshaderpar3[j]=0.0f;
            SHADER[i].pixshaderpar4[j]=0.0f;
            }

         for (j=0; j<SHADERFRGTEXMAX; j++)
            {
            SHADER[i].pixshadertexid[j]=0;
            SHADER[i].pixshadertexw[j]=0;
            SHADER[i].pixshadertexh[j]=0;
            SHADER[i].pixshadertexcl[j]=0;
            SHADER[i].pixshadertexmm[j]=0;
            SHADER[i].pixshadertexnofree[j]=0;
            }
         }

      initsnippets();
      initshader();
      }

   INSTANCES++;
   }

// copy constructor
miniARB::miniARB(const miniARB &arb)
   {INSTANCES++;}

// destructor
miniARB::~miniARB()
   {
   int i,j;

   INSTANCES--;

   if (INSTANCES==0)
      {
      for (i=0; i<SHADERMAX; i++)
         {
         SHADER[i].occupied=0;

         if (SHADER[i].vtxprog!=NULL)
            {
            free(SHADER[i].vtxprog);
            SHADER[i].vtxprog=NULL;
            if (SHADER[i].vtxprogid!=0) deletevtxprog(SHADER[i].vtxprogid);
            SHADER[i].vtxdirty=0;
            }

         if (SHADER[i].frgprog!=NULL)
            {
            free(SHADER[i].frgprog);
            SHADER[i].frgprog=NULL;
            if (SHADER[i].frgprogid!=0) deletefrgprog(SHADER[i].frgprogid);
            SHADER[i].frgdirty=0;
            }

         for (j=0; j<SHADERVTXPRMMAX; j++)
            {
            SHADER[i].vtxshaderpar1[j]=0.0f;
            SHADER[i].vtxshaderpar2[j]=0.0f;
            SHADER[i].vtxshaderpar3[j]=0.0f;
            SHADER[i].vtxshaderpar4[j]=0.0f;
            }

         for (j=0; j<SHADERFRGPRMMAX; j++)
            {
            SHADER[i].pixshaderpar1[j]=0.0f;
            SHADER[i].pixshaderpar2[j]=0.0f;
            SHADER[i].pixshaderpar3[j]=0.0f;
            SHADER[i].pixshaderpar4[j]=0.0f;
            }

         for (j=0; j<SHADERFRGTEXMAX; j++)
            if (SHADER[i].pixshadertexid[j]!=0)
               {
               deletetexmap(SHADER[i].pixshadertexid[j]);
               SHADER[i].pixshadertexid[j]=0;

               SHADER[i].pixshadertexw[j]=0;
               SHADER[i].pixshadertexh[j]=0;
               SHADER[i].pixshadertexcl[j]=0;
               SHADER[i].pixshadertexmm[j]=0;
               SHADER[i].pixshadertexnofree[j]=0;
               }
         }

      freesnippets();
      }
   }

// set vertex shader
void miniARB::setvtxshader(int num,char *vtxprog)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();

   if (SHADER[num].vtxprog!=NULL)
      {
      if (strcmp(vtxprog,SHADER[num].vtxprog)==0) return;
      free(SHADER[num].vtxprog);
      }

   SHADER[num].vtxprog=strdup(vtxprog);
   SHADER[num].vtxdirty=1;
   }

// concatenate vertex shader from snippets
void miniARB::concatvtxshader(int num,const char *snippetname)
   {
   int n;

   char *vtxprog;

   if (num<0 || num>=SHADERMAX) ERRORMSG();

   for (n=0; n<SNIPPETS; n++)
      if (strcmp(snippetname,SNIPPET[n].snippetname)==0)
         {
         vtxprog=strdup2(SHADER[num].vtxprog,SNIPPET[n].snippet);
         if (SHADER[num].vtxprog!=NULL) free(SHADER[num].vtxprog);
         SHADER[num].vtxprog=vtxprog;
         SHADER[num].vtxdirty=1;

         break;
         }
   }

// set vertex shader parameter vector
void miniARB::setvtxshaderparams(int num,float p1,float p2,float p3,float p4,int n)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();
   if (n<0 || n>=SHADERVTXPRMMAX) ERRORMSG();

   SHADER[num].vtxshaderpar1[n]=p1;
   SHADER[num].vtxshaderpar2[n]=p2;
   SHADER[num].vtxshaderpar3[n]=p3;
   SHADER[num].vtxshaderpar4[n]=p4;
   }

// enable vertex shader
void miniARB::enablevtxshader(int num)
   {
   int i;

   if (num<0 || num>=SHADERMAX) ERRORMSG();

   if (SHADER[num].vtxprog!=NULL)
      {
      if (SHADER[num].vtxdirty!=0)
         {
         if (SHADER[num].vtxprogid!=0) deletevtxprog(SHADER[num].vtxprogid);

         SHADER[num].vtxprogid=buildvtxprog(SHADER[num].vtxprog);
         SHADER[num].vtxdirty=0;
         }

      if (SHADER[num].vtxprogid!=0)
         {
         bindvtxprog(SHADER[num].vtxprogid);

         for (i=0; i<SHADERVTXPRMMAX; i++)
            setvtxprogpar(i,SHADER[num].vtxshaderpar1[i],SHADER[num].vtxshaderpar2[i],SHADER[num].vtxshaderpar3[i],SHADER[num].vtxshaderpar4[i]);
         }
      }
   }

// disable vertex shader
void miniARB::disablevtxshader(int num)
   {
   if (SHADER[num].vtxprog!=NULL)
      if (SHADER[num].vtxprogid!=0) bindvtxprog(0);
   }

// set pixel shader
void miniARB::setpixshader(int num,char *frgprog)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();

   if (SHADER[num].frgprog!=NULL)
      {
      if (strcmp(frgprog,SHADER[num].frgprog)==0) return;
      free(SHADER[num].frgprog);
      }

   SHADER[num].frgprog=strdup(frgprog);
   SHADER[num].frgdirty=1;
   }

// concatenate pixel shader from snippets
void miniARB::concatpixshader(int num,const char *snippetname)
   {
   int n;

   char *frgprog;

   if (num<0 || num>=SHADERMAX) ERRORMSG();

   for (n=0; n<SNIPPETS; n++)
      if (strcmp(snippetname,SNIPPET[n].snippetname)==0)
         {
         frgprog=strdup2(SHADER[num].frgprog,SNIPPET[n].snippet);
         if (SHADER[num].frgprog!=NULL) free(SHADER[num].frgprog);
         SHADER[num].frgprog=frgprog;
         SHADER[num].frgdirty=1;

         break;
         }
   }

// set pixel shader parameter vector
void miniARB::setpixshaderparams(int num,float p1,float p2,float p3,float p4,int n)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();
   if (n<0 || n>=SHADERFRGPRMMAX) ERRORMSG();

   SHADER[num].pixshaderpar1[n]=p1;
   SHADER[num].pixshaderpar2[n]=p2;
   SHADER[num].pixshaderpar3[n]=p3;
   SHADER[num].pixshaderpar4[n]=p4;
   }

// set pixel shader RGB texture map
void miniARB::setpixshadertexRGB(int num,unsigned char *image,int width,int height,int clamp,int mipmaps,int n)
   {setpixshadertex(num,image,width,height,3,clamp,mipmaps,n);}

// set pixel shader RGBA texture map
void miniARB::setpixshadertexRGBA(int num,unsigned char *image,int width,int height,int clamp,int mipmaps,int n)
   {setpixshadertex(num,image,width,height,4,clamp,mipmaps,n);}

// set pixel shader RGB[A] texture map
void miniARB::setpixshadertex(int num,unsigned char *image,int width,int height,int components,int clamp,int mipmaps,int n)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();
   if (n<0 || n>=SHADERFRGTEXMAX) ERRORMSG();

   if (width<2 || height<2) ERRORMSG();

   if (SHADER[num].pixshadertexid[n]!=0)
      {
      if (SHADER[num].pixshadertexnofree[n]==0)  deletetexmap(SHADER[num].pixshadertexid[n]);
      SHADER[num].pixshadertexid[n]=0;

      SHADER[num].pixshadertexw[n]=0;
      SHADER[num].pixshadertexh[n]=0;
      SHADER[num].pixshadertexcl[n]=0;
      SHADER[num].pixshadertexmm[n]=0;
      SHADER[num].pixshadertexnofree[n]=0;
      }

   if (image!=NULL)
      {
      if (components==1) SHADER[num].pixshadertexid[n]=buildLtexmap(image,&width,&height,mipmaps);
      else if (components==3) SHADER[num].pixshadertexid[n]=buildRGBtexmap(image,&width,&height,mipmaps);
      else if (components==4) SHADER[num].pixshadertexid[n]=buildRGBAtexmap(image,&width,&height,mipmaps);
      else ERRORMSG();

      SHADER[num].pixshadertexw[n]=width;
      SHADER[num].pixshadertexh[n]=height;
      SHADER[num].pixshadertexcl[n]=clamp;
      SHADER[num].pixshadertexmm[n]=mipmaps;
      SHADER[num].pixshadertexnofree[n]=0;
      }
   }

// set pixel shader RGB[A] texture map from texture id
void miniARB::setpixshadertexid(int num,int texid,int width,int height,int clamp,int mipmaps,int n)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();
   if (n<0 || n>=SHADERFRGTEXMAX) ERRORMSG();

   if (width<2 || height<2) ERRORMSG();

   if (SHADER[num].pixshadertexid[n]!=0)
      {
      if (SHADER[num].pixshadertexnofree[n]==0) deletetexmap(SHADER[num].pixshadertexid[n]);
      SHADER[num].pixshadertexid[n]=0;

      SHADER[num].pixshadertexw[n]=0;
      SHADER[num].pixshadertexh[n]=0;
      SHADER[num].pixshadertexcl[n]=0;
      SHADER[num].pixshadertexmm[n]=0;
      SHADER[num].pixshadertexnofree[n]=0;
      }

   SHADER[num].pixshadertexid[n]=texid;
   SHADER[num].pixshadertexw[n]=width;
   SHADER[num].pixshadertexh[n]=height;
   SHADER[num].pixshadertexcl[n]=clamp;
   SHADER[num].pixshadertexmm[n]=mipmaps;
   SHADER[num].pixshadertexnofree[n]=1;
   }

// set pixel shader RGB[A] texture map from image buffer
void miniARB::setpixshadertexbuf(int num,databuf *buf,int clamp,int mipmaps,int n)
   {
   int texid;
   int width,height;

   if (num<0 || num>=SHADERMAX) ERRORMSG();
   if (n<0 || n>=SHADERFRGTEXMAX) ERRORMSG();

   if (buf->missing())
      {
      WARNMSG("invalid DB object");
      return;
      }

   if (buf->xsize<2 || buf->ysize<2 ||
       buf->zsize>1 || buf->tsteps>1) ERRORMSG();

   width=buf->xsize;
   height=buf->ysize;

   if (SHADER[num].pixshadertexid[n]!=0)
      {
      deletetexmap(SHADER[num].pixshadertexid[n]);
      SHADER[num].pixshadertexid[n]=0;

      SHADER[num].pixshadertexw[n]=0;
      SHADER[num].pixshadertexh[n]=0;
      SHADER[num].pixshadertexcl[n]=0;
      SHADER[num].pixshadertexmm[n]=0;
      SHADER[num].pixshadertexnofree[n]=0;
      }

   texid=db2texid(buf,&width,&height,&mipmaps);

   SHADER[num].pixshadertexid[n]=texid;
   SHADER[num].pixshadertexw[n]=width;
   SHADER[num].pixshadertexh[n]=height;
   SHADER[num].pixshadertexcl[n]=clamp;
   SHADER[num].pixshadertexmm[n]=mipmaps;
   SHADER[num].pixshadertexnofree[n]=0;
   }

// enable pixel shader
void miniARB::enablepixshader(int num)
   {
   int i;

   if (num<0 || num>=SHADERMAX) ERRORMSG();

   if (SHADER[num].frgprog!=NULL)
      {
      if (SHADER[num].frgdirty!=0)
         {
         if (SHADER[num].frgprogid!=0) deletefrgprog(SHADER[num].frgprogid);

         SHADER[num].frgprogid=buildfrgprog(SHADER[num].frgprog);
         SHADER[num].frgdirty=0;
         }

      if (SHADER[num].frgprogid!=0)
         {
         bindfrgprog(SHADER[num].frgprogid);

         for (i=0; i<SHADERFRGPRMMAX; i++)
            setfrgprogpar(i,SHADER[num].pixshaderpar1[i],SHADER[num].pixshaderpar2[i],SHADER[num].pixshaderpar3[i],SHADER[num].pixshaderpar4[i]);

         for (i=0; i<SHADERFRGTEXMAX; i++)
            if (SHADER[num].pixshadertexid[i]!=0)
               {
               texunit(i);

               if (SHADER[num].pixshadertexcl[i]==0) bindtexmap(SHADER[num].pixshadertexid[i],0,0,0,SHADER[num].pixshadertexmm[i]);
               else bindtexmap(SHADER[num].pixshadertexid[i],SHADER[num].pixshadertexw[i],SHADER[num].pixshadertexh[i],0,SHADER[num].pixshadertexmm[i]);
               }

         texunit(0);
         }
      }
   }

// disable pixel shader
void miniARB::disablepixshader(int num)
   {
   int i;

   if (SHADER[num].frgprog!=NULL)
      {
      if (SHADER[num].frgprogid!=0) bindfrgprog(0);

      for (i=0; i<SHADERFRGTEXMAX; i++)
         if (SHADER[num].pixshadertexid[i]!=0)
            {
            texunit(i);
            bindtexmap(0,0,0,0,0);
            }

      texunit(0);
      }
   }

// set direct shading parameters
void miniARB::setshadedirectparams(int num,
                                   float lightdir[3],
                                   float lightbias,float lightoffset)
   {
   setpixshaderparams(num,lightdir[0],lightdir[1],lightdir[2],0.0f,5);
   setpixshaderparams(num,lightbias,lightoffset,0.0f,0.0f,4);
   }

// set global direct shading parameters
void miniARB::setglobalshadedirectparams(float lightdir[3],
                                         float lightbias,float lightoffset)
   {
   global_lightdir[0]=lightdir[0];
   global_lightdir[1]=lightdir[1];
   global_lightdir[2]=lightdir[2];

   global_lightbias=lightbias;
   global_lightoffset=lightoffset;
   }

// set direct texturing parameters
void miniARB::settexturedirectparams(int num,
                                     float lightdir[3],
                                     float transbias,float transoffset)
   {
   setpixshaderparams(num,lightdir[0],lightdir[1],lightdir[2],PI/2,5);
   setpixshaderparams(num,transbias,transoffset,0.0f,0.5f,4);
   }

// set global direct texturing parameters
void miniARB::setglobaltexturedirectparams(float lightdir[3],
                                           float transbias,float transoffset)
   {
   global_lightdir[0]=lightdir[0];
   global_lightdir[1]=lightdir[1];
   global_lightdir[2]=lightdir[2];

   global_transbias=transbias;
   global_transoffset=transoffset;
   }

// set fog parameters
void miniARB::setfogparams(int num,
                           float fogstart,float fogend,
                           float fogdensity,
                           float fogcolor[3])
   {
   float fog_a,fog_b,fog_c;

   // calculate the fog parameters
   if (fogstart<fogend && fogdensity>0.0f)
      {
      fog_a=fsqr(1.0f/fogend);
      fog_b=0.0f;
      fog_c=log(fmin(0.5f*fogdensity,1.0f))/(2.0f*log(fogstart/fogend/2.0f+0.5f));
      }
   else
      {
      fog_a=0.0f;
      fog_b=0.0f;
      fog_c=1.0f;
      }

   // pass the fog parameters
   setpixshaderparams(num,fog_a,fog_b,fog_c,0.0f,6);
   setpixshaderparams(num,fogcolor[0],fogcolor[1],fogcolor[2],0.0f,7);
   }

// set global fog parameters
void miniARB::setglobalfogparams(float fogstart,float fogend,
                                 float fogdensity,
                                 float fogcolor[3])
   {
   global_fogstart=fogstart;
   global_fogend=fogend;

   global_fogdensity=fogdensity;

   global_fogcolor[0]=fogcolor[0];
   global_fogcolor[1]=fogcolor[1];
   global_fogcolor[2]=fogcolor[2];
   }

// set stereo interlacing parameters
void miniARB::setsfxparams(int num,
                           int sfxmode)
   {
   float sfx_a=0.0f,sfx_b=0.0f,sfx_c=0.5f,sfx_d=0.5f;

   if (sfxmode==1) {sfx_a=0.5f; sfx_c=0.5f;}
   else if (sfxmode==2) {sfx_a=0.5f; sfx_c=0.0f;}
   else if (sfxmode==3) {sfx_b=0.5f; sfx_d=0.0f;}
   else if (sfxmode==4) {sfx_b=0.5f; sfx_d=0.5f;}

   setpixshaderparams(num,sfx_a,sfx_b,sfx_c,sfx_d,2);
   }

// set global stereo interlacing parameters
void miniARB::setglobalsfxparams(int sfxmode)
   {global_sfxmode=sfxmode;}

// set global shader
void miniARB::useglobalshader(int num)
   {USEGLOBALSHADER=num;}

// get global shader
int miniARB::getglobalshader()
   {return(USEGLOBALSHADER);}

// get vertex shader
char *miniARB::getvtxshader(int num)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();

   return(SHADER[num].vtxprog);
   }

// get pixel shader
char *miniARB::getpixshader(int num)
   {
   if (num<0 || num>=SHADERMAX) ERRORMSG();

   return(SHADER[num].frgprog);
   }
