// (c) by Stefan Roettger, licensed under LGPL 2.1

#ifndef MINICRYPT_H
#define MINICRYPT_H

#include <stdint.h>
#include <stddef.h>

#include <string>

class minicrypt: public std::string
   {
   public:

   //! ctor
   minicrypt(const std::string &s) : std::string(s) {}
   minicrypt(const char *s,size_t n) : std::string(s,n) {}

   //! Adler-32 hash function
   static uint32_t hash(const std::string &buffer);

   //! CMWC random function
   static void seed(const std::string &key);
   static uint32_t random();

   //! encrypt with key
   void encrypt(const std::string &key);

   //! decrypt with key
   void decrypt(const std::string &key);

   protected:

   void permute_buffer();
   void permute_buffer(size_t s,size_t n);
   void unpermute_buffer();
   void unpermute_buffer(size_t s,size_t n);
   void xor_buffer();

   static uint32_t Q[4096];
   static uint32_t q,p,c;

   std::string reverse(const std::string s);
   };

#endif
