// (c) by Stefan Roettger, licensed under LGPL 2.1

#define OPENGL_TEST // enable this to perform an OpenGL test
#define MINICRS_TEST // enable this to perform a test of the minicrs class
#define MINICOORD_TEST // enable this to perform a test of the minicoord class
#define MINISTRING_TEST // enable this to perform a test of the ministring class
#define MINISTRIP_TEST // enable this to perform a test of the ministrip class
#define MINITIME_TEST // enable this to perform a test of the minitime module
#define MINIPATH_TEST // enable this to perform a test of the minipath module
#define MINIXML_TEST // enable this to perform a test of the minixml module
#define GLSLMATH_TEST // enable this to perform a test of the glslmath module
#define LUNAFUNCTOR_TEST // enable this to perform a test of the lunafunctor module
#define MERGESORT_TEST // enable this to perform a test of the mergesort module
#define MINI3D_TEST // enable this to perform a test of the mini3D module
#define MINIWLS_TEST // enable this to perform a test of the miniWLS module
#define MINIWLS3D_TEST // enable this to perform a test of the miniWLS3D module
#define MINICRYPT_TEST // enable this to perform a test of the minicrypt module

#include <mini/minibase.h>
#include <mini/miniOGL.h>
#include <mini/minicrs.h>
#include <mini/minicoord.h>
#include <mini/ministring.h>
#include <mini/ministrip.h>
#include <mini/minitime.h>
#include <mini/minipath.h>
#include <mini/minixml.h>
#include <mini/glslmath.h>
#include <mini/lunafunctor.h>
#include <mini/minisort.h>
#include <mini/mini3D.h>
#include <mini/miniclod.h>
#include <mini/miniwls.h>
#include <mini/minicrypt.h>
#include <mini/miniio.h>

#ifndef __APPLE__
#include <GL/glut.h>
#else
#include <GLUT/glut.h>
#endif

static int winwidth,winheight,winid;

BOOLINT equal(double x,double y,double e=1E-5)
   {return(fabs(x-y)<e);}

void displayfunc()
   {
   float ex=0.0f,ey=10.0f,ez=30.0f;
   float dx=0.0f,dy=-0.25f,dz=-1.0f;
   float ux=0.0f,uy=1.0f,uz=0.0f;
   float fovy=60.0f;
   float aspect=1.0f;
   float nearp=1.0f;
   float farp=100.0f;

   aspect=(float)winwidth/winheight;

   glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(fovy,aspect,nearp,farp);

   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   gluLookAt(ex,ey,ez,ex+dx,ey+dy,ez+dz,ux,uy,uz);

   // add OpenGL test code here:
   // ...

   beginfans();
   color(1,0,0);
   beginfan();
   fanvertex(-5,3,0);
   fanvertex(5,3,0);
   fanvertex(5,0,5);
   fanvertex(-5,0,5);
   endfans();

   // end of test OpenGL code

   glutSwapBuffers();
   }

void reshapefunc(int width,int height)
   {
   winwidth=width;
   winheight=height;

   glViewport(0,0,width,height);

   displayfunc();
   }

void keyboardfunc(unsigned char key,int x,int y)
   {
   float mousex,mousey;

   mousex=(float)x/(winwidth-1);
   mousey=(float)y/(winwidth-1);

   if (mousex<0.0f) mousex=0.0f;
   else if (mousex>1.0f) mousex=1.0f;

   if (mousey<0.0f) mousey=0.0f;
   else if (mousey>1.0f) mousey=1.0f;

   if (key=='q' || key==27)
      {
      glutDestroyWindow(winid);
      exit(0);
      }
   }

int main(int argc,char *argv[])
   {
   int failure=0;

   if (argc!=1)
      {
      printf("usage: %s\n",argv[0]);
      exit(1);
      }

#ifdef OPENGL_TEST
   winwidth=winheight=512;

   glutInit(&argc,argv);
   glutInitWindowSize(winwidth,winheight);
   glutInitDisplayMode(GLUT_RGB|GLUT_ALPHA|GLUT_DEPTH|GLUT_DOUBLE);
   winid=glutCreateWindow("libMini Test");

   glutDisplayFunc(displayfunc);
   glutReshapeFunc(reshapefunc);
   glutMouseFunc(NULL);
   glutMotionFunc(NULL);
   glutKeyboardFunc(keyboardfunc);
   glutSpecialFunc(NULL);
   glutIdleFunc(NULL);
#endif

   printf("running test code\n\n");

   // add test code here:
   // ...

#ifdef OPENGL_TEST
   {
   std::cout << "miniOGL:" << std::endl;
   miniOGL::print_unsupported_glexts();
   miniOGL::print_graphics_info();
   if (get_unsupported_glexts()==0) std::cout << "OPENGL SUCCESS" << std::endl;
   else {std::cout << "OPENGL FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINICRS_TEST
   {
   std::cout << "minicrs:" << std::endl;
   int z0=4;
   double lon=minicrs::UTMZ2L(z0);
   std::cout << z0 << " -> " << lon/3600 << std::endl;
   int z=minicrs::LL2UTMZ(0,lon);
   if (z==z0) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINICOORD_TEST
   {
   std::cout << "minicoord:" << std::endl;
   minicoord c1(miniv4d(11*3600,49*3600,0),minicoord::MINICOORD_LLH,0,minicoord::MINICOORD_DATUM_WGS84);
   std::cout << c1 << std::endl;
   minicoord c2=c1;
   c2.convert2(minicoord::MINICOORD_UTM);
   std::cout << " " << c2 << std::endl;
   std::cout << " " << c2.to_string() << std::endl;
   ministring info=c2.to_string();
   minicoord c3;
   c3.from_string(info);
   std::cout << " " << c3 << std::endl;
   c3.convert2llh();
   std::cout << c3 << std::endl;
   if (c1.getdist(c3)<1E-2) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }

   {
   std::cout << "minicoord:" << std::endl;
   minicoord u(miniv4d(500000,0,0),minicoord::MINICOORD_UTM,4,minicoord::MINICOORD_DATUM_WGS84);
   std::cout << u << std::endl;
   u.convert2(minicoord::MINICOORD_LLH);
   std::cout << " " << u << std::endl;
   u.convert2(minicoord::MINICOORD_UTM);
   std::cout << " " << u << std::endl;
   if (u.vec.x==500000) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINISTRING_TEST
   {
   std::cout << "ministring:" << std::endl;
   ministring test1("this is a simple test string");
   ministring test2("this is a s1mple test string");
   ministring check1=test1.checksum(4);
   ministring check2=test2.checksum(4);
   std::cout << test1 << " -> " << check1 << std::endl;
   std::cout << test2 << " -> " << check2 << std::endl;
   unsigned int count=0;
   for (unsigned int i=0; i<check1.getsize(); i++) if (check1[i]!=check2[i]) count++;
   if (count==1) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINISTRIP_TEST
   {
   std::cout << "ministrip:" << std::endl;
   ministrip strip1(3,3);
   strip1.setcol(1,0,0);
   strip1.setnrm(0,0,-1);
   strip1.beginstrip();
   strip1.addvtx(0,0,0);
   strip1.addvtx(1,0,0);
   strip1.addvtx(0,1,0);
   strip1.addvtx(1,1,0);
   strip1.beginstrip();
   strip1.addvtx(0,0,0);
   strip1.addvtx(-1,0,0);
   strip1.addvtx(0,1,0);
   strip1.addvtx(-1,1,0);
   int size1=strip1.getsize();
   ministring strip=strip1.to_string();
   unsigned int before=strip.size();
   miniv3d bboxmin1,bboxmax1;
   strip1.getbbox(bboxmin1,bboxmax1);
   std::cout << "ministrip: size1=" << size1 << " before=" << before << std::endl;
   std::cout << "ministrip: bbox1=" << bboxmin1 << "->" << bboxmax1 << std::endl;
   std::cout << "ministrip: " << strip << std::endl;
   ministrip strip2;
   strip2.from_string(strip);
   int size2=strip2.getsize();
   unsigned int after=strip.size();
   miniv3d bboxmin2,bboxmax2;
   strip2.getbbox(bboxmin2,bboxmax2);
   std::cout << "ministrip: size2=" << size2 << " after=" << after << std::endl;
   std::cout << "ministrip: bbox2=" << bboxmin2 << "->" << bboxmax2 << std::endl;
   if (size1==10 && size2==10 && before==453 && after==0 &&
       bboxmin1==bboxmin2 && bboxmax1==bboxmax2)
      std::cout << "SUCCESS" << std::endl;
   else
      {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINITIME_TEST
   {
   std::cout << "minitime:" << std::endl;
   double s1971=utc2minitime(1971,7,2,14,34,0,0);
   double s1970=utc2minitime(1970,1,1,0,0,0,0);
   double s1969=utc2minitime(1969,7,1,0,0,0,700);
   double s2014=utc2minitime(2014,3,19,15,18,14,0);
   ministring utc("2014-03-13T10:33:09.876Z");
   std::cout << "minitime: s1971=" << s1971 << std::endl;
   std::cout << "minitime: s1970=" << s1970 << std::endl;
   double sutc=utc2unixtime(2014,3,13,10,33,9,876);
   double utcd=utc2unixtime(utc)-sutc;
   if ((long long)(s1971+0.5)==0 &&
       (long long)(s1970+0.5)==-47313239 &&
       (long long)(s1969+0.5)==-63210838 &&
       (long long)(s2014+0.5)==1346978654 &&
       utcd==0.0)
      std::cout << "SUCCESS" << std::endl;
   else
      {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINIPATH_TEST
   {
   std::cout << "minipath:" << std::endl;
   minicoord point1(miniv4d(11*3600,49*3600,0),minicoord::MINICOORD_LLH,0,minicoord::MINICOORD_DATUM_WGS84);
   minicoord point2(miniv4d(500000,0,0),minicoord::MINICOORD_UTM,4,minicoord::MINICOORD_DATUM_WGS84);
   minipath path;
   path.append(point1);
   path.append(point2);
   path.set_constraints();
   double length=path.get_length();
   std::cout << "minipath: length=" << length << std::endl;
   if (dabs(length-1.15704e+07)<1E1) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINIXML_TEST
   {
   std::cout << "minixml:" << std::endl;
   ministrings txt;
   txt.from_string("<?xml version=\"1.0\"?> \n <tag><subtag>text</subtag></tag> \n <anothertag answer=42> < / anothertag >","\n");
   minixml xml;
   xml.from_strings(txt);
   if (txt.empty()) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef GLSLMATH_TEST
   {
   std::cout << "glslmath:" << std::endl;
   mat3 M(vec3(0,1,0),vec3(-1,0,0),vec3(0,0,1));
   mat4 D(vec4(1,-1,1.01010101,1));
   std::cout << M << std::endl;
   glslmath::print(M);
   std::cout << D << std::endl;
   glslmath::print(D);
   if (glslmath::test()==0) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef LUNAFUNCTOR_TEST
   {
   std::cout << "lunafunctor:" << std::endl;
   lunafunctor functor;
   functor.expr("1-2*x");
   if (functor.f(0.5f)==0.0f) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MERGESORT_TEST
   {
   std::cout << "mergesort:" << std::endl;
   int a=3,b=2,c=1;
   minidyna<int *> array;
   array.push_back(&a);
   array.push_back(&b);
   array.push_back(&c);
   mergesort<int>(array);
   if (*(array[0])==1 && *(array[1])==2 && *(array[2])==3) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINI3D_TEST
   {
   std::cout << "mini3D:" << std::endl;
   minipath path;
   path.unset_constraints();
   if (!path.load("test.csv")) path.load("test/test.csv");
   double length=path.get_length()/1000;
   miniCLODcontainer clod;
   clod.set(path);
   minicoord eye;
   eye.set_llh(49.9,10.9,0);
   eye.convert2ecef();
   vec3 e=eye.getpos();
   clod.create(e);
   unsigned int points=clod.getPoints()->size();
   mini3Dcounter t;
   t.line(*clod.getPoints());
   t.band(*clod.getPoints());
   minicoord lookat=path[1];
   lookat.convert2ecef();
   vec3 l=lookat.getpos();
   vec3 u=l;
   mat4 p=mat4::perspective(90,1,1E2,1E5);
   mat4 mv=mat4::lookat(e,l,u);
   mat4 mvp=p*mv;
   t.postMultiply(mvp);
   t.render();
   unsigned int lines=t.numLines();
   unsigned int triangles=t.numTriangles();
   double llength=t.lineLength();
   double tarea=t.triangleArea();
   if (dabs(length-10410.5)<1E1 && points==4 &&
       lines==3 && triangles==9 &&
       dabs(llength-484.1)<1E1 && dabs(tarea-1850.0)<1E1)
      std::cout << "SUCCESS" << std::endl;
   else
      {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINIWLS_TEST
   {
   std::cout << "miniWLS:" << std::endl;
   WLS_Point p1={-1,1,1};
   WLS_Point p2={0,1,1};
   WLS_Point p3={1,3,1};
   miniWLS wls;
   wls.push_back(p1);
   wls.quadratic_fit();
   vec3 c1=wls.coefficients();
   wls.push_back(p2);
   wls.quadratic_fit();
   vec3 c2=wls.coefficients();
   wls.push_back(p3);
   wls.quadratic_fit();
   vec3 c3=wls.coefficients();
   if (c1.x==1 && c1.y==0 && c1.z==0 &&
       c2.x==1 && c2.y==0 && c2.z==0 &&
       c3.x==1 && c3.y==1 && c3.z==1)
      std::cout << "SUCCESS" << std::endl;
   else
      {std::cout << "FAILURE" << std::endl; failure++;}
   }
   {
   std::cout << "miniWLSA:" << std::endl;
   WLS_Point a1={0,350,1};
   WLS_Point a2={1,360,1};
   WLS_Point a3={2,20,1};
   WLS_Point a4={3,40,1};
   miniWLSA wlsa;
   wlsa.push_back(a1);
   wlsa.push_back(a2);
   wlsa.push_back(a3);
   wlsa.push_back(a4);
   wlsa.linear_fit();
   if (dabs(wlsa.eval(3)-38)<1E-9) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINIWLS3D_TEST
   {
   std::cout << "miniWLS3D:" << std::endl;
   WLS_Point3D p1={vec3(0,0,0),0,1};
   WLS_Point3D p2={vec3(1,1,1),1,1};
   WLS_Point3D p3={vec3(2,2,2),2,1};
   WLS_Point3D p4={vec3(3,3,2),3,1};
   WLS_Point3D p5={vec3(4,4,1),4,1};
   WLS_Point3D p6={vec3(5,5,0),5,1};
   miniWLS3D wls3D;
   wls3D.push_back(p1);
   wls3D.linear_fit();
   vec3 a1,b1;
   wls3D.coefficients(a1,b1);
   wls3D.push_back(p2);
   wls3D.linear_fit();
   vec3 a2,b2;
   wls3D.coefficients(a2,b2);
   wls3D.push_back(p3);
   wls3D.linear_fit();
   vec3 a3,b3;
   wls3D.coefficients(a3,b3);
   wls3D.push_back(p4);
   wls3D.push_back(p5);
   wls3D.push_back(p6);
   wls3D.linear_fit();
   vec3 a4,b4;
   wls3D.coefficients(a4,b4);
   wls3D.parabolic_fit();
   vec3 a5,b5,c5;
   wls3D.coefficients(a5,b5,c5);
   if (a1.x==0 && a1.y==0 && a1.z==0 &&
       b1.x==0 && b1.y==0 && b1.z==0 &&
       a2.x==0 && a2.y==0 && a2.z==0 &&
       b2.x==1 && b2.y==1 && b2.z==1 &&
       a3.x==0 && a3.y==0 && a3.z==0 &&
       b3.x==1 && b3.y==1 && b3.z==1 &&
       a4.x==0 && a4.y==0 && a4.z==1 &&
       b4.x==1 && b4.y==1 && b4.z==0 &&
       a5.x==0 && a5.y==0 && equal(a5.z,-0.0714286) &&
       b5.x==1 && b5.y==1 && equal(b5.z,1.60714) &&
       c5.x==0 && c5.y==0 && equal(c5.z,-0.321429))
      std::cout << "SUCCESS" << std::endl;
   else
      {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

#ifdef MINICRYPT_TEST
   {
   std::cout << "minicrypt:" << std::endl;
   int errors=0;
   minicrypt test("just a little test");
   test.encrypt("test key");
   test.decrypt("test key");
   if (test!="just a little test") errors++;
   long long bytes;
   unsigned char *data=miniio::readfile("data/SkyDome.ppm",&bytes);
   if (data)
      {
      minicrypt test((char *)data,bytes);
      test.encrypt("another test key");
      test.decrypt("another test key");
      if (test!=std::string((char *)data,bytes)) errors++;
      free(data);
      }
   if (errors==0) std::cout << "SUCCESS" << std::endl;
   else {std::cout << "FAILURE" << std::endl; failure++;}
   }
#endif

   // end of test code

   printf("\nfinished test code\n");

#ifdef OPENGLTEST
   glutMainLoop();
#endif

   return(failure);
   }
