// (c) by Stefan Roettger, licensed under GPL 3.0

#ifndef SWIPEFILTER_H
#define SWIPEFILTER_H

#ifdef HAVE_QT5
#include <QtWidgets>
#else
#include <QtGui>
#endif

//! swiping direction
enum SwipeDirection
{
   SwipeNone,
   SwipeLeft,
   SwipeRight,
   SwipeDown,
   SwipeUp
};

//! swiping event filter
class SwipeFilter: public QObject
{
   Q_OBJECT

public:

   SwipeFilter(QWidget *parent);
   virtual ~SwipeFilter();

   void set_delta(int delta = mouse_delta);
   void set_damping(double damping = motion_damping);
   void set_doubleswipe(int delta = doubleswipe_time);

protected:

   bool eventFilter(QObject *obj, QEvent *event);

   QWidget *parent_;

   bool leftButtonDown_;
   QWidget *widgetHit_;
   QPoint lastPos_;
   int delta_;

   SwipeDirection direction_;
   int offset_;
   QTime time_;
   double elapsed_;
   double kinetic_;

   QTimer motion_;
   double damping_;

   QTime firstSwipe_;
   SwipeDirection lastSwipe_;
   int doubleswipe_;

protected slots:

   void motion();

signals:

   void click(QWidget *widget);
   void doubleClick(QWidget *widget);
   void swipe(SwipeDirection direction, double speed);
   void move(SwipeDirection direction, int offset);
   void kinetic(SwipeDirection direction, int offset);
   void doubleswipe(SwipeDirection direction);

private:

   static const int mouse_delta;
   static const double mouse_threshold;

   static const double motion_fps;
   static const double motion_damping;

   static const int doubleswipe_time;
};

#endif
